@@ -1,29 +1,35 @@
 import re
-import zipfile
 from collections.abc import Iterable
 from dataclasses import dataclass
 from pathlib import Path
-from typing import IO
+from typing import TypedDict
 from xml.etree import ElementTree
 from xml.etree.ElementTree import Element
 
 import yaml
 
-from . import __version__, util
+from anyerplint import preprocessor
+from anyerplint.library import Library
+from anyerplint.translate import Translator
+
+from . import __version__, tnex, util
 
 ROOT = Path(__file__).absolute().parent
 
 
-@dataclass
-class Library:
-    variables: set[str]
-    loopvariables: set[str]
-    calls: set[str]
-    functions: set[str]
-    jumps: set[str]
-    tags: set[str]
-    # if true, accumulate everything found to library
-    teaching: bool
+ReportData = TypedDict(
+    "ReportData",
+    {
+        "Unknown variables": list[str],
+        "Unknown calls": list[str],
+        "Unknown jumps": list[str],
+        "Unknown tags": list[str],
+        "Other errors": list[str],
+        "Expression errors": list[dict[str, str | list[str]]],
+        "Unknown loop variables": list[str],
+    },
+    total=False,
+)
 
 
 # ruff lead this alone man
@@ -34,14 +40,14 @@ def do_check(
     libs: list[str],
     targets: list[str],
     teaching: bool,
+    *,
     nostdlib: bool = False,
-) -> dict[str, str | dict[str, list[str]]]:
+) -> dict[str, str | ReportData]:
     emit(f"AnyErpLintVersion: {__version__.strip()}")
 
     lib_vars = Library(
         variables=set(),
         calls=set(),
-        functions=set(),
         jumps=set(),
         tags=set(),
         teaching=teaching,
@@ -52,13 +58,13 @@ def do_check(
 
     if not nostdlib:
         local_app_data = util.get_app_local_dir()
-        feed_lib(lib_vars, local_app_data)
+        lib_vars.load_from_disk(local_app_data)
 
     for lib in libs:
-        feed_lib(lib_vars, Path(lib))
+        lib_vars.load_from_disk(Path(lib))
 
     has_errors = False
-    all_errors: dict[str, str | dict[str, list[str]]] = {}
+    all_errors: dict[str, str | ReportData] = {}
     for target in targets:
         files = util.glob_plus(target)
         errs = check_files(lib_vars, files)
@@ -67,12 +73,17 @@ def do_check(
             has_errors = True
 
     if lib_vars.teaching:
-        write_lib(lib_vars)
+        stdlib_path = util.get_app_local_dir()
+        lib_vars.save_to_disk(stdlib_path, emit)
 
     if has_errors:
         emit("Errors found: >")
         error_counts = {
-            f: (sum(len(l) for l in e.values()) if not isinstance(e, str) else -1)
+            f: (
+                sum(len(prop) for prop in e.values() if isinstance(prop, list))
+                if not isinstance(e, str)
+                else -1
+            )
             for f, e in all_errors.items()
         }
         rep = sorted((k, v) for (k, v) in error_counts.items())
@@ -82,83 +93,6 @@ def do_check(
     return all_errors
 
 
-def write_lib(lib_vars: Library) -> None:
-    def write_file(fname: Path, lines: list[str]) -> None:
-        emit("  - ", fname)
-        f = open(fname, "wb")
-        for line in lines:
-            try:
-                if not line.strip().isascii():
-                    continue
-                enc = line.strip().encode()
-            except UnicodeEncodeError:
-                # skip bad lines for now
-                continue
-            f.write(enc)
-            f.write(b"\n")
-
-    emit("Writing found function to:")
-
-    stdlib = util.get_app_local_dir()
-    calls_file = stdlib / "builtin_calls.txt"
-    write_file(
-        calls_file,
-        sorted(call for call in lib_vars.calls if is_valid_call(call)),
-    )
-
-    functions_file = stdlib / "builtin_functions.txt"
-    write_file(functions_file, sorted(lib_vars.functions))
-
-    jumps_file = stdlib / "builtin_jumps.txt"
-    write_file(jumps_file, sorted(lib_vars.jumps))
-
-    tags_file = stdlib / "builtin_tags.txt"
-    write_file(tags_file, sorted(lib_vars.tags))
-
-    vars_file = stdlib / "builtin_vars.txt"
-    write_file(vars_file, sorted(lib_vars.variables))
-
-    loopvars_file = stdlib / "builtin_loopvars.txt"
-    write_file(loopvars_file, sorted({lv.lower() for lv in lib_vars.loopvariables}))
-
-
-def is_valid_call(call: str) -> bool:
-    return True
-
-
-def feed_lib(lib_vars: Library, libdir: Path) -> None:
-    def feed_set(set: set[str], fobj: IO[bytes]) -> None:
-        set.update(line.decode().strip() for line in fobj.readlines())
-
-    def visit_file(fname: str, fobj: IO[bytes]) -> None:
-        if fname.endswith("_calls.txt"):
-            feed_set(lib_vars.calls, fobj)
-        elif fname.endswith("_functions.txt"):
-            feed_set(lib_vars.functions, fobj)
-        elif fname.endswith("_jumps.txt"):
-            feed_set(lib_vars.jumps, fobj)
-        elif fname.endswith("_tags.txt"):
-            feed_set(lib_vars.tags, fobj)
-        elif fname.endswith("_vars.txt"):
-            feed_set(lib_vars.variables, fobj)
-        elif fname.endswith("_loopvars.txt"):
-            feed_set(lib_vars.loopvariables, fobj)
-            lib_vars.loopvariables = {lv.lower() for lv in lib_vars.loopvariables}
-
-    if not libdir.exists():
-        return
-
-    # files on file system
-    for p in libdir.glob("*_*.txt"):
-        visit_file(str(p), p.open("rb"))
-
-    # files in all the zip files
-    for f in libdir.glob("*.zip"):
-        zf = zipfile.ZipFile(f, "r")
-        for zn in zf.namelist():
-            visit_file(zn, zf.open(zn, "r"))
-
-
 def report_fatal(fname: Path, ex: Exception) -> str:
     message = f"FATAL, {ex}"
     emit(f"{fname}: {message}")
@@ -168,18 +102,22 @@ def report_fatal(fname: Path, ex: Exception) -> str:
 def should_skip_file(fpath: Path) -> bool:
     if fpath.suffix != ".xml":
         return True
-    with util.open_text_file(fpath) as f:
-        text = f.read(10000)
-        return not ("<erpConnector" in text or "<section" in text)
+    text = util.read_text_file(fpath)
+    return not ("<erpConnector" in text or "<section" in text)
 
 
-def check_file(lib_vars: Library, file: Path) -> str | dict[str, list[str]] | None:
+def check_file(lib_vars: Library, file: Path) -> str | ReportData | None:
     try:
         if should_skip_file(file):
             emit(f"{file}: SKIP nontemplate")
             return None
-        r = parse_file(file, lib_vars.teaching)
-    except (ElementTree.ParseError, PermissionError, UnicodeDecodeError) as e:
+        r = parse_file(file, teaching=lib_vars.teaching)
+    except (
+        ElementTree.ParseError,
+        PermissionError,
+        UnicodeDecodeError,
+        preprocessor.PreprocessError,
+    ) as e:
         return report_fatal(file, e)
 
     return report(lib_vars, str(file), r)
@@ -187,8 +125,8 @@ def check_file(lib_vars: Library, file: Path) -> str | dict[str, list[str]] | No
 
 def check_files(
     lib_vars: Library, files: Iterable[Path]
-) -> dict[str, str | dict[str, list[str]]]:
-    errs: dict[str, str | dict[str, list[str]]] = {}
+) -> dict[str, str | ReportData]:
+    errs: dict[str, str | ReportData] = {}
     for f in files:
         errlist = check_file(lib_vars, f)
         if errlist:
@@ -197,8 +135,23 @@ def check_files(
     return errs
 
 
+@dataclass
+class ExprError:
+    """Represents an error in an expression."""
+
+    summary: str
+    translated: str
+    messages: list[str]
+
+    def render(self) -> dict[str, str | list[str]]:
+        """Render yaml compatible object."""
+        return {"Error": self.summary, "Simple": self.translated, "Msg": self.messages}
+
+
 @dataclass
 class Parsed:
+    """The things found while parsing a file."""
+
     var_decl: set[str]
     var_used: set[str]
     alltags: set[str]
@@ -207,11 +160,11 @@ class Parsed:
     syntax_errors: list[str]
     loop_var_decl: set[str]
     loop_var_use: set[str]
-    used_functions: set[str]
+    expression_errors: list[ExprError]
     full_content: str
 
 
-def report(lib_vars: Library, fname: str, p: Parsed) -> dict[str, list[str]]:
+def report(lib_vars: Library, fname: str, p: Parsed) -> ReportData:
     undeclared_vars = p.var_used - p.var_decl
     undeclared_vars.difference_update(lib_vars.variables)
     unknown_loop_variables = {
@@ -223,13 +176,12 @@ def report(lib_vars: Library, fname: str, p: Parsed) -> dict[str, list[str]]:
 
     if lib_vars.teaching:
         lib_vars.calls.update(p.calls)
-        lib_vars.functions.update(f for f in p.used_functions if f.isupper())
         lib_vars.jumps.update(p.jumps)
         lib_vars.tags.update(p.alltags)
         lib_vars.variables.update(undeclared_vars)
         lib_vars.loopvariables.update(lv.lower() for lv in unknown_loop_variables)
 
-    errors = {}
+    errors: ReportData = {}
 
     if undeclared_vars:
         errors["Unknown variables"] = util.add_linenumbers(
@@ -242,14 +194,6 @@ def report(lib_vars: Library, fname: str, p: Parsed) -> dict[str, list[str]]:
     if unknown_calls:
         errors["Unknown calls"] = sorted(unknown_calls)
 
-    unknown_functions = p.used_functions
-    unknown_functions.difference_update(lib_vars.functions)
-    if unknown_functions:
-        errors["Unknown functions"] = util.add_linenumbers(
-            p.full_content,
-            sorted(unknown_functions),
-        )
-
     unknown_jumps = p.jumps
     unknown_jumps.difference_update(lib_vars.jumps)
     if unknown_jumps:
@@ -263,16 +207,20 @@ def report(lib_vars: Library, fname: str, p: Parsed) -> dict[str, list[str]]:
     if p.syntax_errors:
         errors["Other errors"] = list(p.syntax_errors)
 
+    if p.expression_errors:
+        errors["Expression errors"] = [e.render() for e in p.expression_errors]
+
     if unknown_loop_variables:
         errors["Unknown loop variables"] = sorted(unknown_loop_variables)
 
     if errors:
-        emit(yaml.dump({fname: errors}, width=200).strip())
+        emit(yaml.dump({fname: errors}, width=200, sort_keys=False).strip())
 
     return errors
 
 
 key_params: dict[str, str] = {
+    "bw_exec": "command",
     "bw_file_functions": "command",
     "bw_table_method": "command",
     "bw_string_functions": "operation",
@@ -297,10 +245,10 @@ def summarize_call(node: Element) -> str:
 
 
 def summarize_output(node: Element) -> str:
-    type = node.attrib.get("type", "NOTYPE")
+    output_type = node.attrib.get("type", "NOTYPE")
     params = {p.attrib.get("name", "NONAME") for p in node.iter("parameter")}
 
-    return ("Output " + type + " - " + ",".join(sorted(params))).strip()
+    return ("Output " + output_type + " - " + ",".join(sorted(params))).strip()
 
 
 def summarize_tag(node: Element) -> str:
@@ -347,17 +295,13 @@ def brace_check(s: str) -> list[str]:
         if flush_stack:
             stack = []
     if stack:
-        pretty_stack = ", ".join(f"{ch} {l}" for ch, l in stack)
+        pretty_stack = ", ".join(f"{ch} {lnum}" for ch, lnum in stack)
         errors.append(
             f"File ended with mismatched braces, remaining in stack (char, linenum): {pretty_stack}",
         )
     return errors
 
 
-# xxx not really needed due to new logic
-MAGIC_VAR_NAMES = {"error", "return", "response", "invoice.i"}
-
-
 def describe_node(n: Element) -> str:
     return "<" + n.tag + str(n.attrib) + ">"
 
@@ -440,28 +384,81 @@ def check_outputresources(root: Element) -> list[str]:
     return errors
 
 
-def find_invalid_expressions(content: str) -> list[str]:
-    expr_regex = r"{(.*?)}"
+def full_parse_expression(expr: str) -> tuple[str, list[str]]:
+    try:
+        parsed = tnex.parse(expr, expand_entities=True)
+        if not parsed:
+            return "", [f"Empty parse result from: {expr}"]
+        translator = Translator()
+        translated = translator.translate(parsed)
+    except tnex.ParseError as e:
+        return "", [f"Fatal parse error: {e}"]
+    else:
+        return translated, translator.errors
 
-    def is_invalid_expression(expr: re.Match[str]) -> bool:
-        if "," not in expr.group(1):
-            # Regular expressions can contain things like {3}
-            if expr.group(1).isnumeric():
-                return False
-            return True
-        return False
 
-    return [
-        f"Invalid expression {expr.group(0)}, line {linenum}"
-        for linenum, line in enumerate(content.splitlines(), 1)
-        for expr in re.finditer(expr_regex, line)
-        if is_invalid_expression(expr)
+MAX_EXPRESSION_LINES = 500
+
+
+def find_invalid_expressions(expressions: list[tuple[str, int]]) -> list[ExprError]:
+    res: list[ExprError] = []
+    for expr, linenum in expressions:
+        if expr and expr[0].isalpha() and "," not in expr:
+            res.append(
+                ExprError(
+                    summary=f"Line {linenum}, expression contains no comma: {{{expr}}}",
+                    translated="N/A",
+                    messages=[],
+                )
+            )
+            continue
+        if expr.count("\n") > MAX_EXPRESSION_LINES:
+            res.append(
+                ExprError(
+                    summary=f"Line {linenum}, expression too long, likely unterminated: {{{expr[:200]}}}",
+                    translated="N/A",
+                    messages=[],
+                )
+            )
+            continue
+
+        translated, errors = full_parse_expression(expr)
+
+        if errors:
+            res.append(
+                ExprError(
+                    summary=f"Line {linenum}, expression: {{{expr}}}",
+                    translated=translated,
+                    messages=errors,
+                )
+            )
+    return res
+
+
+def find_invalid_text_in_tags(tree: ElementTree.ElementTree) -> list[str]:
+    no_text_allowed_tags = [
+        "sections",
+        "section",
+        "method",
+        "output",
+        "outputCommands",
+        "builtInMethodParameterList",
     ]
 
+    errors: list[str] = []
+    for notext in no_text_allowed_tags:
+        nodes = tree.iter(notext)
+        errors.extend(
+            f"Node should not contain text: {describe_node(n)} -- {n.text.strip()}"
+            for n in nodes
+            if n.text and n.text.strip()
+        )
+    return errors
 
-def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
+
+def parse_file(fname: Path, *, teaching: bool = False) -> Parsed:
     tree = ElementTree.parse(fname)
-    raw_cont = util.open_text_file(fname).read()
+    raw_cont = util.read_text_file(fname)
     comments_removed = util.replace_commented_xml_with_empty_lines(raw_cont)
 
     vardecl = {
@@ -473,22 +470,19 @@ def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
         for v in tree.iter("parameter")
     }
 
+    expressions = preprocessor.get_expressions(comments_removed)
+
     prop_regex = r"([a-zA-Z_][a-zA-Z0-9_.]*),(\w+)"
 
     propaccess = {
         (match.group(1), match.group(2))
-        for match in re.finditer(prop_regex, comments_removed)
+        for expr, line in expressions
+        for match in re.finditer(prop_regex, expr)
     }
     varuse = {name for expr_type, name in propaccess if expr_type.lower() == "v"}
-    used_functions = {
-        expr_type + "," + name
-        for expr_type, name in propaccess
-        if expr_type.lower() == "f"
-    }
 
     # what to do with p params?
     otherpropaccess = {k for k, v in propaccess if k.lower() not in ["v", "f", "p"]}
-    otherpropaccess.difference_update(MAGIC_VAR_NAMES)
     calls = {summarize_call(v) for v in tree.iter("builtInMethodParameterList")}
     outputs = {summarize_output(v) for v in tree.iter("output")}
     calls.update(outputs)
@@ -507,29 +501,13 @@ def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
     }
 
     errors = []
-
+    expr_errors = []
     if not teaching:
         cdata_removed = util.replace_cdata_with_empty_lines(comments_removed)
         errors.extend(brace_check(cdata_removed))
 
-        no_text_allowed_tags = [
-            "sections",
-            "section",
-            "method",
-            "output",
-            "outputCommands",
-            "builtInMethodParameterList",
-        ]
-        for notext in no_text_allowed_tags:
-            nodes = tree.iter(notext)
-            for n in nodes:
-                if n and n.text and n.text.strip():
-                    errors.append(
-                        "Node should not contain text: "
-                        + describe_node(n)
-                        + " -- "
-                        + n.text.strip(),
-                    )
+        errors.extend(find_invalid_text_in_tags(tree))
+
         var_passwords = {v for v in vardecl if is_illegal_password(v, vardecl[v])}
         param_passwords = {
             p for p in all_params if is_illegal_password(p, all_params[p])
@@ -541,8 +519,7 @@ def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
         outputresource_errors = check_outputresources(tree.getroot())
         errors.extend(outputresource_errors)
 
-        expr_errors = find_invalid_expressions(comments_removed)
-        errors.extend(expr_errors)
+        expr_errors = find_invalid_expressions(expressions)
 
     return Parsed(
         var_decl=set(vardecl),
@@ -550,9 +527,9 @@ def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
         alltags=alltags,
         calls=calls,
         jumps=jumps,
-        used_functions=used_functions,
         syntax_errors=errors,
         loop_var_decl=loop_data_sources,
         loop_var_use=otherpropaccess,
         full_content=comments_removed,
+        expression_errors=expr_errors,
     )
