@@ -1,4 +1,5 @@
 import argparse
+import sys
 
 from anyerplint import business_logic
 
@@ -8,7 +9,9 @@ def handle_check(args: argparse.Namespace) -> None:
 
     Don't put business logic here. Only parse the arguments and pass forward.
     """
-    business_logic.do_check(args.lib, args.target, args.teach)
+    errors = business_logic.do_check(args.lib, args.target, args.teach)
+    if args.exitcode and errors:
+        sys.exit(1)
 
 
 def init_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
@@ -29,5 +32,9 @@ def init_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
         help="Run in teaching mode (update function list)",
     )
 
+    parser.add_argument(
+        "--exitcode", action="store_true", help="Set nonzero exit code on error"
+    )
+
     parser.set_defaults(func=handle_check)
     return parser
